import ctypes

# ctypes.wintypes is available on linux? might not be in older versions
# so not touching for now
# various Mock that allows the doc to be build on darwin/linux
try:
    import ctypes.wintypes
except ValueError:
    from unittest.mock import Mock

    ctypes.wintypes = Mock()
    ctypes.wintypes.MAX_PATH = 1
    ctypes.WinDLL = Mock()

# Allows docs to be built with ci tools that run on linux
try:
    getattr(ctypes, "WinDLL")
except AttributeError:
    class MockObject:
        def __init__(self, *args, **kwargs):
            return

        def __call__(self, *args, **kwargs):
            return self

        def __getattr__(self, item):
            return self

    ctypes.Windll = MockObject()



import ctypes
import struct

import pymem.memory
import pymem.exception


class RemotePointer(object):
    """Pointer capable of reading the value mapped into another process memory.

    Parameters
    ----------
    handle: int
        Handle to the process
    v: int, RemotePointer, any ctypes type
        The address value
    endianess: str
        The endianess of the remote pointer, defaulting to little-endian

    Raises
    ------
    PymemAlignmentError
        If endianess is not a valid alignment

    Notes
    -----
    The bool of RemotePointer checks if the internal value is 0
    """

    ALIGNMENTS = {
        'little-endian': '<',
        'big-endian': '>'
    }

    def __init__(self, handle, v, endianess='little-endian'):
        self._set_value(v)

        if endianess not in RemotePointer.ALIGNMENTS:
            # TODO: maybe make this a ValueError in next major version
            raise pymem.exception.PymemAlignmentError(
                "{endianess} is not a valid alignment, it should be one from: {alignments}".format(**{
                    'endianess': endianess,
                    'alignments': ', '.join(RemotePointer.ALIGNMENTS.keys())
                })
            )
        self.endianess = endianess

        self.handle = handle
        self._memory_value = None

    def __bool__(self):
        return bool(self.value)

    def _set_value(self, v):
        """Given a v value will set up the internal kitchen to map internal v to the correct
        type. self.v has to be a ctype instance.
        """
        if isinstance(v, RemotePointer):
            self.v = v.cvalue
        elif isinstance(v, int) and not hasattr(v, 'value'):
            if v > 2147483647:
                self.v = ctypes.c_ulonglong(v)
            else:
                self.v = ctypes.c_uint(v)
        elif isinstance(v, ctypes._SimpleCData):
            self.v = v
        else:
            raise pymem.exception.PymemTypeError(
                "{type} is not an allowed type, it should be one from: {allowed_types}".format(**{
                    'type': 'None' if not v else str(type(v)),
                    'allowed_types': ', '.join([
                        'RemotePointer', 'ctypes', 'int'
                    ])
                }))

    def __add__(self, a):
        self._memory_value = self.value + a
        return self.cvalue

    @property
    def value(self):
        """Reads targeted process memory and returns the value pointed by the given address.

        Returns
        -------
        int
            The value pointed at by this remote pointer
        """
        if self._memory_value:
            return self._memory_value
        content = pymem.memory.read_bytes(
            self.handle, self.v.value, struct.calcsize(self.v._type_)
        )
        fmt = '{alignment}{type}'.format(**{
            'alignment': RemotePointer.ALIGNMENTS[self.endianess],
            'type': self.v._type_
        })
        content = struct.unpack(fmt, content)
        self._memory_value = content[0]
        return self._memory_value

    @property
    def cvalue(self):
        """Reads targeted process memory and returns the value pointed by the given address.

        Returns
        -------
        a ctypes type
            The value pointed at by this remote pointer as a ctypes type instance
        """
        v = self.v.__class__(self.value)
        return v



import sys

import pymem.memory
import pymem.ressources.kernel32
import pymem.ressources.structure

try:
    # faster than builtin re
    import regex as re
except ImportError:
    import re


# TODO: warn that pattern is a regex and may need to be escaped
# TODO: 2.0 rename to pattern_scan_page
def scan_pattern_page(handle, address, pattern, *, return_multiple=False):
    """Search a byte pattern given a memory location.
    Will query memory location information and search over until it reaches the
    length of the memory page. If nothing is found the function returns the
    next page location.

    Parameters
    ----------
    handle: int
        Handle to an open object
    address: int
        An address to search from
    pattern: bytes
        A regex byte pattern to search for
    return_multiple: bool
        If multiple results should be returned instead of stopping on the first

    Returns
    -------
    tuple
        next_region, found address

        found address may be None if one was not found, or we didn't have permission to scan
        the region

        if return_multiple is True found address will instead be a list of found addresses
        or an empty list if no results

    Examples
    --------
    >>> pm = pymem.Pymem("Notepad.exe")
    >>> address_reference = 0x7ABC00001
    # Here the "." means that the byte can be any byte; a "wildcard"
    # also note that this pattern may be outdated
    >>> bytes_pattern = b".\\x00\\x00\\x00\\x01\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00" \\
    ...                 b"\\x00\\x00\\x00\\x00\\x00\\x00..\\x00\\x00..\\x00\\x00\\x64\\x04"
    >>> character_count_address = pymem.pattern.scan_pattern_page(pm.process_handle, address_reference, bytes_pattern)
    """
    mbi = pymem.memory.virtual_query(handle, address)
    next_region = mbi.BaseAddress + mbi.RegionSize
    allowed_protections = [
        pymem.ressources.structure.MEMORY_PROTECTION.PAGE_EXECUTE_READ,
        pymem.ressources.structure.MEMORY_PROTECTION.PAGE_EXECUTE_READWRITE,
        pymem.ressources.structure.MEMORY_PROTECTION.PAGE_READWRITE,
        pymem.ressources.structure.MEMORY_PROTECTION.PAGE_READONLY,
    ]
    if mbi.state != pymem.ressources.structure.MEMORY_STATE.MEM_COMMIT or mbi.protect not in allowed_protections:
        return next_region, None

    page_bytes = pymem.memory.read_bytes(handle, address, mbi.RegionSize)

    if not return_multiple:
        found = None
        match = re.search(pattern, page_bytes, re.DOTALL)

        if match:
            found = address + match.span()[0]

    else:
        found = []

        for match in re.finditer(pattern, page_bytes, re.DOTALL):
            found_address = address + match.span()[0]
            found.append(found_address)

    return next_region, found


def pattern_scan_module(handle, module, pattern, *, return_multiple=False):
    """Given a handle over an opened process and a module will scan memory after
    a byte pattern and return its corresponding memory address.

    Parameters
    ----------
    handle: int
        Handle to an open object
    module: MODULEINFO
        An instance of a given module
    pattern: bytes
        A regex byte pattern to search for
    return_multiple: bool
        If multiple results should be returned instead of stopping on the first

    Returns
    -------
    int, list, optional
        Memory address of given pattern, or None if one was not found
        or a list of found addresses in return_multiple is True

    Examples
    --------
    >>> pm = pymem.Pymem("Notepad.exe")
    # Here the "." means that the byte can be any byte; a "wildcard"
    # also note that this pattern may be outdated
    >>> bytes_pattern = b".\\x00\\x00\\x00\\x01\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00" \\
    ...                 b"\\x00\\x00\\x00\\x00\\x00\\x00..\\x00\\x00..\\x00\\x00\\x64\\x04"
    >>> module_notepad = pymem.process.module_from_name(pm.process_handle, "Notepad.exe")
    >>> character_count_address = pymem.pattern.pattern_scan_module(pm.process_handle, module_notepad, bytes_pattern)
    """
    base_address = module.lpBaseOfDll
    max_address = module.lpBaseOfDll + module.SizeOfImage
    page_address = base_address

    if not return_multiple:
        found = None
        while page_address < max_address:
            page_address, found = scan_pattern_page(handle, page_address, pattern)

            if found:
                break

    else:
        found = []
        while page_address < max_address:
            page_address, new_found = scan_pattern_page(handle, page_address, pattern, return_multiple=True)

            if new_found:
                found += new_found

    return found


def pattern_scan_all(handle, pattern, *, return_multiple=False):
    """Scan the entire address space for a given regex pattern

    Parameters
    ----------
    handle: int
        Handle to an open process
    pattern: bytes
        A regex bytes pattern to search for
    return_multiple: bool
        If multiple results should be returned

    Returns
    -------
    int, list, optional
        Memory address of given pattern, or None if one was not found
        or a list of found addresses in return_multiple is True
    """
    next_region = 0

    found = []
    user_space_limit = 0x7FFFFFFF0000 if sys.maxsize > 2**32 else 0x7fff0000
    while next_region < user_space_limit:
        next_region, page_found = scan_pattern_page(
            handle,
            next_region,
            pattern,
            return_multiple=return_multiple
        )

        if not return_multiple and page_found:
            return page_found

        if page_found:
            found += page_found

    if not return_multiple:
        return None

    return found



import ctypes

import pymem.ressources.structure

try:
    dll = ctypes.WinDLL('advapi32.dll')
except AttributeError:
    class MockObject:
        def __getattr__(self, item):
            return self

    dll = MockObject()

#: The LookupPrivilegeValue function retrieves the locally unique identifier (LUID) used on a specified system to
#: locally represent the specified privilege name.
#:
#: https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-lookupprivilegevaluea
LookupPrivilegeValue = dll.LookupPrivilegeValueW
LookupPrivilegeValue.argtypes = (
    ctypes.c_wchar_p,  # system name
    ctypes.c_wchar_p,  # name
    ctypes.POINTER(pymem.ressources.structure.LUID),
)
LookupPrivilegeValue.restype = ctypes.c_long

#: The LookupPrivilegeName function retrieves the name that corresponds to the privilege represented on a specific
#: system by a specified locally unique identifier (LUID).
#:
#: https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-lookupprivilegenamea
LookupPrivilegeName = dll.LookupPrivilegeNameW
LookupPrivilegeName.argtypes = (
    ctypes.c_wchar_p,  # lpSystemName
    ctypes.POINTER(pymem.ressources.structure.LUID),  # lpLuid
    ctypes.c_wchar_p,  # lpName
    ctypes.POINTER(ctypes.c_ulong),  # cchName
)
LookupPrivilegeName.restype = ctypes.c_long


#: The OpenProcessToken function opens the access token associated with a process.
#:
#: https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocesstoken
OpenProcessToken = dll.OpenProcessToken
OpenProcessToken.argtypes = (
    ctypes.c_void_p,
    ctypes.c_ulong,
    ctypes.POINTER(ctypes.c_void_p)
)
OpenProcessToken.restype = ctypes.c_long


#: The AdjustTokenPrivileges function enables or disables privileges in the specified access token.
#: Enabling or disabling privileges in an access token requires TOKEN_ADJUST_PRIVILEGES access.
#:
#: https://docs.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-adjusttokenprivileges
AdjustTokenPrivileges = dll.AdjustTokenPrivileges
AdjustTokenPrivileges.restype = ctypes.c_long
AdjustTokenPrivileges.argtypes = (
    ctypes.c_void_p,  # TokenHandle
    ctypes.c_long,  # DisableAllPrivileges
    pymem.ressources.structure.PTOKEN_PRIVILEGES,  # NewState (optional)
    ctypes.c_ulong,  # BufferLength of PreviousState
    pymem.ressources.structure.PTOKEN_PRIVILEGES,  # PreviousState (out, optional)
    ctypes.POINTER(ctypes.c_ulong),  # ReturnLength
)



import ctypes
import locale
import logging
import os

import pymem.ressources.advapi32
import pymem.ressources.kernel32
import pymem.ressources.psapi
import pymem.ressources.structure


logger = logging.getLogger(__name__)


def get_python_dll(version):
    """Given a python dll version will find its path using the current process as a placeholder

    Parameters
    ----------
    version: str
        A string representation of python version as a dll (python38.dll)

    Returns
    -------
    str
        The full path of dll
    """
    current_process_id = os.getpid()
    current_process_handle = pymem.process.open(current_process_id)
    for module in pymem.process.enum_process_module(current_process_handle):
        if module.name == version:
            return module.filename


def inject_dll(handle, filepath):
    """Inject a dll into opened process.

    Parameters
    ----------
    handle: int
        Handle to an open object
    filepath: bytes
        Dll to be injected filepath

    Returns
    -------
    DWORD
        The address of injected dll
    """
    filepath_address = pymem.ressources.kernel32.VirtualAllocEx(
        handle,
        0,
        len(filepath),
        pymem.ressources.structure.MEMORY_STATE.MEM_COMMIT.value | pymem.ressources.structure.MEMORY_STATE.MEM_RESERVE.value,
        pymem.ressources.structure.MEMORY_PROTECTION.PAGE_EXECUTE_READWRITE.value
    )
    pymem.ressources.kernel32.WriteProcessMemory(handle, filepath_address, filepath, len(filepath), None)
    kernel32_handle = pymem.ressources.kernel32.GetModuleHandleW("kernel32.dll")
    load_library_a_address = pymem.ressources.kernel32.GetProcAddress(kernel32_handle, b"LoadLibraryA")
    thread_h = pymem.ressources.kernel32.CreateRemoteThread(
        handle, None, 0, load_library_a_address, filepath_address, 0, None
    )
    pymem.ressources.kernel32.WaitForSingleObject(thread_h, -1)
    pymem.ressources.kernel32.VirtualFreeEx(
        handle, filepath_address, len(filepath), pymem.ressources.structure.MEMORY_STATE.MEM_RELEASE.value
    )
    dll_name = os.path.basename(filepath)
    dll_name = dll_name.decode('ascii')
    module_address = pymem.ressources.kernel32.GetModuleHandleW(dll_name)
    return module_address


def get_luid(name):
    """Get the LUID for the SeCreateSymbolicLinkPrivilege
    """
    luid = pymem.ressources.structure.LUID()
    res = pymem.ressources.advapi32.LookupPrivilegeValue(None, name, luid)
    if not res > 0:
        raise RuntimeError("Couldn't lookup privilege value")
    return luid


def get_process_token():
    """Get the current process token
    """
    token = ctypes.c_void_p()
    res = pymem.ressources.advapi32.OpenProcessToken(
        ctypes.windll.kernel32.GetCurrentProcess(),
        pymem.ressources.structure.TOKEN.TOKEN_ALL_ACCESS,
        token
    )
    if not res > 0:
        raise RuntimeError("Couldn't get process token")
    return token


def set_debug_privilege(lpszPrivilege, bEnablePrivilege):
    """Leverage current process privileges.

    Parameters
    ----------
    lpszPrivilege: str
        Privilege name
    bEnablePrivilege: bool
        Enable privilege

    Returns
    -------
    bool
        If privileges have been leveraged
    """
    # create a space in memory for a TOKEN_PRIVILEGES structure
    #  with one element
    size = ctypes.sizeof(pymem.ressources.structure.TOKEN_PRIVILEGES)
    size += ctypes.sizeof(pymem.ressources.structure.LUID_AND_ATTRIBUTES)
    buffer = ctypes.create_string_buffer(size)

    tp = ctypes.cast(buffer, ctypes.POINTER(pymem.ressources.structure.TOKEN_PRIVILEGES)).contents
    tp.count = 1
    tp.get_array()[0].LUID = get_luid(lpszPrivilege)
    tp.get_array()[0].Attributes = (
        pymem.ressources.structure.SE_TOKEN_PRIVILEGE.SE_PRIVILEGE_ENABLED if bEnablePrivilege else 0
    )
    token = get_process_token()
    res = pymem.ressources.advapi32.AdjustTokenPrivileges(token, False, tp, 0, None, None)
    if res == 0:
        raise RuntimeError("AdjustTokenPrivileges error: 0x%08x\n" % ctypes.GetLastError())

    ERROR_NOT_ALL_ASSIGNED = 1300
    return ctypes.windll.kernel32.GetLastError() != ERROR_NOT_ALL_ASSIGNED


def base_module(handle):
    """Returns process base module

    Parameters
    ----------
    handle: int
        A valid handle to an open object

    Returns
    -------
    MODULEINFO
        The base module of the process
    """
    hModules = (ctypes.c_void_p * 1024)()
    process_module_success = pymem.ressources.psapi.EnumProcessModulesEx(
        handle,
        ctypes.byref(hModules),
        ctypes.sizeof(hModules),
        ctypes.byref(ctypes.c_ulong()),
        pymem.ressources.structure.EnumProcessModuleEX.LIST_MODULES_ALL
    )
    if not process_module_success:
        return  # xxx
    module_info = pymem.ressources.structure.MODULEINFO(handle)
    pymem.ressources.psapi.GetModuleInformation(
        handle,
        ctypes.c_void_p(hModules[0]),
        ctypes.byref(module_info),
        ctypes.sizeof(module_info)
    )
    return module_info


def open(process_id, debug=True, process_access=None):
    """Open a process given its process_id.
    By default, the process is opened with full access and in debug mode.

    https://msdn.microsoft.com/en-us/library/windows/desktop/ms684320%28v=vs.85%29.aspx
    https://msdn.microsoft.com/en-us/library/windows/desktop/aa379588%28v=vs.85%29.aspx

    Parameters
    ----------
    process_id: int
        The identifier of the process to be opened
    debug: bool
        If the process should be opened in debug mode
    process_access: pymem.ressources.structure.PROCESS
        Desired access level, defaulting to all access

    Returns
    -------
    int
        A handle to the opened process
    """
    if not process_access:
        process_access = pymem.ressources.structure.PROCESS.PROCESS_ALL_ACCESS.value
    if debug:
        set_debug_privilege('SeDebugPrivilege', True)
    process_handle = pymem.ressources.kernel32.OpenProcess(process_access, False, process_id)
    return process_handle


def open_main_thread(process_id):
    """List given process threads and return a handle to first created one.

    Parameters
    ----------
    process_id: int
        The identifier of the process

    Returns
    -------
    int
        A handle to the main thread
    """
    threads = enum_process_thread(process_id)
    threads = sorted(threads, key=lambda t32: t32.creation_time)

    if not threads:
        return  # todo: raise exception

    main_thread = threads[0]
    thread_handle = open_thread(main_thread.th32ThreadID)
    return thread_handle


# TODO: impl enum for thread access levels
def open_thread(thread_id, thread_access=None):
    """Opens an existing thread object.
    https://msdn.microsoft.com/en-us/library/windows/desktop/ms684335%28v=vs.85%29.aspx

    Parameters
    ----------
    thread_id: int
        The identifier of the thread to be opened
    thread_access: int
        Desired access level, defaulting to all access

    Returns
    -------
    int
        A handle to the opened thread
    """
    if thread_access is None:
        thread_access = 0x001F03FF
    thread_handle = pymem.ressources.kernel32.OpenThread(thread_access, 0, thread_id)
    return thread_handle


def close_handle(handle):
    """Closes an open object handle.
    https://msdn.microsoft.com/en-us/library/windows/desktop/ms724211%28v=vs.85%29.aspx

    Parameters
    ----------
    handle: int
        A valid handle to an open object

    Returns
    -------
    bool
        If the closure succeeded
    """
    if not handle:
        return
    success = pymem.ressources.kernel32.CloseHandle(handle)
    return success != 0


def list_processes():
    """List all processes
    https://msdn.microsoft.com/en-us/library/windows/desktop/ms682489%28v=vs.85%29.aspx
    https://msdn.microsoft.com/en-us/library/windows/desktop/ms684834%28v=vs.85%29.aspx

    Returns
    -------
    list[ProcessEntry32]
        A list of open process entries
    """
    SNAPPROCESS = 0x00000002
    hSnap = pymem.ressources.kernel32.CreateToolhelp32Snapshot(SNAPPROCESS, 0)
    process_entry = pymem.ressources.structure.ProcessEntry32()
    process_entry.dwSize = ctypes.sizeof(process_entry)
    p32 = pymem.ressources.kernel32.Process32First(hSnap, ctypes.byref(process_entry))
    if p32:
        yield process_entry
    while p32:
        yield process_entry
        p32 = pymem.ressources.kernel32.Process32Next(hSnap, ctypes.byref(process_entry))
    pymem.ressources.kernel32.CloseHandle(hSnap)


def process_from_name(name):
    """Open a process given its name.

    Parameters
    ----------
    name: str
        The name of the process to be opened

    Returns
    -------
    ProcessEntry32
        The process entry of the opened process
    """
    name = name.lower()
    processes = list_processes()
    for process in processes:
        if name in process.szExeFile.decode(locale.getpreferredencoding()).lower():
            return process


def process_from_id(process_id):
    """Open a process given its name.

    Parameters
    ----------
    process_id: int
        The identifier of the process to be opened

    Returns
    -------
    ProcessEntry32
        The process entry of the opened process
    """
    processes = list_processes()
    for process in processes:
        if process_id == process.th32ProcessID:
            return process


def module_from_name(process_handle, module_name):
    """Retrieve a module loaded by given process.

    Parameters
    ----------
    process_handle: int
        Handle to the process to get the module from
    module_name: str
        Name of the module to get

    Returns
    -------
    MODULEINFO
        The retrieved module

    Examples
    --------
    >>> d3d9 = module_from_name(process_handle, 'd3d9')
    """
    module_name = module_name.lower()
    modules = enum_process_module(process_handle)
    for module in modules:
        if module.name.lower() == module_name:
            return module


def enum_process_thread(process_id):
    """List all threads of given processes_id

    Parameters
    ----------
    process_id: int
        Identifier of the process to enum the threads of

    Returns
    -------
    list[ThreadEntry32]
        The process's threads
    """
    TH32CS_SNAPTHREAD = 0x00000004
    hSnap = pymem.ressources.kernel32.CreateToolhelp32Snapshot(TH32CS_SNAPTHREAD, 0)
    thread_entry = pymem.ressources.structure.ThreadEntry32()
    ret = pymem.ressources.kernel32.Thread32First(hSnap, ctypes.byref(thread_entry))

    if not ret:
        raise pymem.exception.PymemError('Could not get Thread32First')

    while ret:
        if thread_entry.th32OwnerProcessID == process_id:
            yield thread_entry
        ret = pymem.ressources.kernel32.Thread32Next(hSnap, ctypes.byref(thread_entry))
    pymem.ressources.kernel32.CloseHandle(hSnap)


def enum_process_module(handle):
    """List and retrieves the base names of the specified loaded module within a process
    https://msdn.microsoft.com/en-us/library/windows/desktop/ms682633(v=vs.85).aspx
    https://msdn.microsoft.com/en-us/library/windows/desktop/ms683196(v=vs.85).aspx

    Parameters
    ----------
    handle: int
        Handle of the process to enum the modules of

    Returns
    -------
    list[MODULEINFO]
        The process's modules
    """
    hModules = (ctypes.c_void_p * 1024)()
    process_module_success = pymem.ressources.psapi.EnumProcessModulesEx(
        handle,
        ctypes.byref(hModules),
        ctypes.sizeof(hModules),
        ctypes.byref(ctypes.c_ulong()),
        pymem.ressources.structure.EnumProcessModuleEX.LIST_MODULES_ALL
    )
    if process_module_success:
        hModules = iter(m for m in hModules if m)
        for hModule in hModules:
            module_info = pymem.ressources.structure.MODULEINFO(handle)
            pymem.ressources.psapi.GetModuleInformation(
                handle,
                ctypes.c_void_p(hModule),
                ctypes.byref(module_info),
                ctypes.sizeof(module_info)
            )
            yield module_info


# TODO: should this be named is_wow64?
def is_64_bit(handle):
    """Determines whether the specified process is running under WOW64 (emulation).

    Parameters
    ----------
    handle: int
        Handle of the process to check wow64 status of

    Returns
    -------
    bool
        If the process is running under wow64
    """
    Wow64Process = ctypes.c_long()
    pymem.ressources.kernel32.IsWow64Process(handle, ctypes.byref(Wow64Process))
    return bool(Wow64Process.value)
